module RadixTree
where

import           Data.Char
import           Data.Maybe
import qualified Data.List as L

import           Data.Map(Map)
import qualified Data.Map as M

import           Data.Set(Set)
import qualified Data.Set as S

-- ------------------------------------------------------------
--

type Key	= String
type Attr	= Set Int

data Table	= Leaf		String Attr
		| Entry		Table  Attr
		| Switch	( Map Char Table )
		| Single	Char Table
		| Empty
		  deriving (Show)

-- ------------------------------------------------------------

emptyTable	:: Table
emptyTable	= Empty

-- ------------------------------------------------------------
--
-- 11 cases

search		:: String -> Table -> Maybe Attr

search k Empty			= Nothing

search k (Leaf rest attr)
    | k == rest			= Just attr
    | otherwise			= Nothing

search "" (Entry table attr)	= Just attr
search k  (Entry table attr)	= search k table

search ""       (Switch sw)	= Nothing
search (c:rest) (Switch sw)	= switch (M.lookup c sw)
    where
    switch Nothing	= Nothing
    switch (Just tab)	= search rest tab

search ""       (Single c' tab)	= Nothing
search (c:rest)	(Single c' tab)
    | c == c'			= search rest tab
    | otherwise			= Nothing

-- ------------------------------------------------------------
--
-- very fast

searchPrefix	:: String -> Table -> Table

searchPrefix "" tab			= tab

searchPrefix k Empty			= emptyTable

searchPrefix k (Entry tab' a')		= searchPrefix k tab'

searchPrefix k (Leaf k' a')
    | L.isPrefixOf k k'			= Leaf (drop (length k) k') a'
    | otherwise				= emptyTable

searchPrefix (c:rest) (Single c' tab')
    | c == c'				= searchPrefix rest tab'
    | otherwise				= emptyTable

searchPrefix (c:rest) (Switch sw')
    | M.member c sw'			= searchPrefix rest
                                            (fromJust $ M.lookup c sw')
    | otherwise				= emptyTable

-- ------------------------------------------------------------
--
-- 14 cases:	completeness, correctness ???

insert	:: String -> Attr -> Table -> Table

insert "" a Empty			= Leaf "" a
insert "" a (Leaf "" a')		= Leaf "" (a' `S.union` a)
insert "" a (Entry tab' a')		= Entry tab' (a' `S.union` a)
insert "" a tab				= Entry tab a

insert k a Empty			= Leaf k a

insert k        a (Leaf k'         a')
    | k == k'				= Leaf k' (a' `S.union` a)

insert k        a (Leaf ""         a')	= Entry (insert k a emptyTable) a'

insert (c:rest) a (Leaf (c':rest') a')
    | c == c'				= Single c (insert rest a (insert rest' a' emptyTable))
    | otherwise				= Switch (M.insert c (insert rest a emptyTable)
						  (M.singleton c' (insert rest' a' emptyTable)))

insert k        a (Entry tab' a')	= Entry (insert k a tab') a'

insert (c:rest) a (Switch sw')
    | M.member c sw'			= Switch (M.adjust (insert rest a) c sw')
    | otherwise				= Switch (M.insert c (insert rest a emptyTable) sw')

insert (c:rest) a (Single c' tab')
    | c == c'				= Single c (insert rest a tab')
    | otherwise				= Switch (M.insert c (insert rest a emptyTable)
						  (M.singleton c' tab'))

-- ------------------------------------------------------------

keys	:: Table -> [String]

keys Empty		= []
keys (Leaf w a)		= [w]
keys (Entry tab a)	= "" : keys tab
keys (Single c tab)	= [ c : w | w <- keys tab ]
keys (Switch sw)	= [ c : w | c <- M.keys sw
                                  , w <- keys (fromJust $ M.lookup c sw)
                          ]

-- ------------------------------------------------------------
--
-- invariant: consistency test
--
-- complete, o.k., insert correct ???
--
-- 1) Empty only legal for empty table
--
-- 2) Switch at least with 2 entries
--
-- 3) no nested Entry constructors
--
-- 4) no Entry with following Leaf with empty string
--
-- 5) all subtables consistent

invTable	:: Table -> Bool
invTable Empty			= True
invTable t			= invTable' t
    where
    invTable' Empty		= False
    invTable' (Switch sw')	= M.size sw' >= 2
				  &&
				  all invTable' (M.elems sw')
    invTable' (Single c t')	= invTable' t'
    invTable' (Entry t' a')	= ( case t' of
				    Entry _ _	-> False
				    Leaf "" _	-> False
				    _		-> True
				  )
			          && invTable' t'
    invTable' (Leaf k' a')	= True

-- ------------------------------------------------------------

tableSpace		:: Table -> Int
tableSpace Empty
    = 0				-- Singleton

tableSpace (Entry t a)
    = 2 + tableSpace t		-- 1 (constructor) + 1 (Table) + 0 (Attr not counted)

tableSpace (Switch m)
    = 2 + 2 * M.size m		-- 1 (constructor) + 1 (Map) +  per Entry (Char, Table)
      + (sum . map tableSpace . M.elems) m

tableSpace (Single c t)
    = 3 + tableSpace t		-- 1 (constructor) + 1 (Char) + 1 (Table)

tableSpace (Leaf s a)
    = 1 + length s		-- 1 (constructor) + length key + 0 (Attr not counted)


tableSize		:: Table -> (Int, Int)
tableSize t
    = (length ks, sum . map length $ ks)
    where
    ks = keys t

-- ------------------------------------------------------------
