module RadixTree2
where

import           Data.Char
import           Data.Maybe

import           Data.Map(Map)
import qualified Data.Map as M

import           Data.Set(Set)
import qualified Data.Set as S

-- ------------------------------------------------------------
--

type Key	= String
type Attr	= Set Int

data Table	= Entry	Table  Attr
		| Switch	( Map Char Table )
		  deriving (Show)

-- ------------------------------------------------------------

emptyTable	:: Table
emptyTable	= Switch M.empty

-- ------------------------------------------------------------
--
-- 5 cases	(completeness, correctness !!!)

search		:: String -> Table -> Maybe Attr

search "" (Entry table attr)	= Just attr

search "" (Switch sw)		= Nothing

search k  (Entry table attr)	= search k table

search k  (Switch sw)		= switch (M.lookup (head k) sw)
    where
    switch Nothing	= Nothing
    switch (Just tab)	= search (tail k) tab

-- ------------------------------------------------------------
--
-- 4 cases, very fast

searchPrefix	:: String -> Table -> Table

searchPrefix "" tab			= tab

searchPrefix k (Entry tab' a')		= searchPrefix k tab'

searchPrefix (c:rest) (Switch sw')
    | M.member c sw'			= searchPrefix rest
                                          (fromJust $ M.lookup c sw')
    | otherwise				= emptyTable

-- ------------------------------------------------------------
--
-- 5 cases	(completeness, correctness !!!)

insert	:: String -> Attr -> Table -> Table


insert ""       a (Entry tab' a')	= Entry tab' (a' `S.union` a)

insert ""       a tab			= Entry tab a

insert k	a (Entry tab' a')	= Entry (insert k a tab') a'

insert (c:rest) a (Switch sw')
    | M.member c sw'			= Switch (M.adjust (insert rest a) c sw')
    | otherwise				= Switch (M.insert c (insert rest a emptyTable) sw')

-- ------------------------------------------------------------

keys	:: Table -> [String]

keys (Entry tab a)	= "" : keys tab
keys (Switch sw)	= [ c : w | c <- M.keys sw
                                  , w <- keys (fromJust $ M.lookup c sw)
                          ]

-- ------------------------------------------------------------
--
-- invariant:
--
-- no nested Entry constructors
--
-- no subtable of a switch is empty

invTable	:: Table -> Bool
invTable (Entry tab attr)
    = case tab of
      Entry _ _  -> False
      Switch _	 -> invTable tab

invTable (Switch sw)
    = all invTable' (M.elems sw)
      where
      invTable' t'		= not (isEmptyTable t')
			  	&&
			  	invTable t'

      isEmptyTable (Entry _ _)	= False
      isEmptyTable (Switch sw)	= M.null sw

-- ------------------------------------------------------------

tableSpace		:: Table -> Int
tableSpace (Entry t a)
    = 2 + tableSpace t		-- 1 (constructor) + 1 (Table) + 0 (Attr not counted)

tableSpace (Switch m)
    = 2 + 2 * M.size m		-- 1 (constructor) + 1 (Map) +  per Entry (Char, Table)
      + (sum . map tableSpace . M.elems) m

tableSize		:: Table -> (Int, Int)
tableSize t
    = (length ks, sum . map length $ ks)
    where
    ks = keys t

-- ------------------------------------------------------------
