module RadixTree3
where

import           Data.Maybe

import           Data.Map(Map)
import qualified Data.Map as M

import           Data.Set(Set)
import qualified Data.Set as S

-- ------------------------------------------------------------
--

type Key	= String
type Attr	= Set Int

data Table	= Switch (Maybe Attr) (Map Char Table)
		  deriving (Show)

-- ------------------------------------------------------------

emptyTable	:: Table
emptyTable	= Switch Nothing M.empty

-- ------------------------------------------------------------
--
-- 2 cases	(completeness, correctness !!!)

search		:: String -> Table -> Maybe Attr

search "" (Switch attr sw)
    = attr

search k  (Switch attr sw)
    = do
      tab <- M.lookup (head k) sw
      search (tail k) tab

-- ------------------------------------------------------------
--
-- 2 cases, very fast

searchPrefix	:: String -> Table -> Table

searchPrefix "" tab	= tab

searchPrefix (c:rest) (Switch attr' sw')
    | M.member c sw'	= searchPrefix rest
                          (fromJust $ M.lookup c sw')
    | otherwise		= emptyTable

-- ------------------------------------------------------------
--
-- 3 cases	(completeness, correctness !!!)

insert	:: String -> Attr -> Table -> Table


insert ""       a (Switch attr' tab')	= Switch (merge attr') tab'
    where
    merge Nothing	= Just a
    merge (Just a')	= Just (a' `S.union` a)

insert (c:rest) a (Switch attr' sw')
    | M.member c sw'	= Switch attr' (M.adjust   (insert rest a) c          sw')
    | otherwise		= Switch attr' (M.insert c (insert rest a emptyTable) sw')

-- ------------------------------------------------------------

keys	:: Table -> [String]

keys (Switch attr sw)	= maybe [] (const [""]) attr
			  ++
			  [ c : w | c <- M.keys sw
                                  , w <- keys (fromJust $ M.lookup c sw)
                          ]

-- ------------------------------------------------------------
--
-- invariant:
--
-- no subtable of a switch is empty
--
-- a switch is empty, if no attribute is stored in the node
-- and the switch map does not contain any entries

invTable	:: Table -> Bool

invTable (Switch attr sw)
    = all invTable' (M.elems sw)
      where

      invTable' t'
	  = not (isEmptyTable t')
	    &&
	    invTable t'

      isEmptyTable (Switch a' sw')
	  = isNothing a'
	    &&
	    M.null sw'

-- ------------------------------------------------------------

tableSpace		:: Table -> Int
tableSpace (Switch e m)
    = 3	                		-- 1 (constructor) + 1 (Maybe) + 1 (Map)
      + 2 * M.size m			-- 2 per Entry (Char, Table)
      + maybe 0 (const 1) e		-- 1 for Just attr
      + (sum . map tableSpace . M.elems) m

tableSize		:: Table -> (Int, Int)
tableSize t
    = (length ks, sum . map length $ ks)
    where
    ks = keys t

-- ------------------------------------------------------------
